%Author: Marco A. Montes de Oca
%Simple implementation of the penalty method for 
%equality constraints using a trust-region-method

clear all; 
more off;

%Dogleg Parameters
mindelta = 1e-04;
delta = 1;
maxiter = 1000;
etag = 0.9;
etaa = 0.1;
gammae = 2;
gammas = 0.5;
gtol=1e-03;
mtol=1e-07;
hessdelta=1e-06;

%Penalty Parameters
mu=1; 
mum=2;
divFlag=0;
citer=100;
pmaxiter=50;

%Initialization
xinit=[1 0.5];
x=xinit;

piter=0;

while piter < pmaxiter 

  %Reset dogleg parameters
 delta=1;

  [current,g,H] = L(x,mu);
  best=current;
  solution = x;
  gradlength = norm(g); 

%    Traj=x;
  Obj=[best];
%    Delta=[delta];

  %Trust-region loop
  iter = 0;
  while iter < citer && gradlength > gtol


%  disp(['Iteration =' num2str(iter) ' Obj. function value = ' num2str(best) ' Current point = ' num2str(x) ]);

  %      %Indefinite Hessian treatment
	[R,pd]=chol(H);
	if pd==1 
	  H=H+(abs(min(eig(H)))+hessdelta)*eye(2);      
	end
      
      %Step computation
      %Newton's step
      N = (-inv(H)*g')';
      if norm(N) <= delta
	p = N;
	%disp(['N'])
      else 
	%Exact steepest descent step
	S = -(gradlength^2/(g*(H*g'))).*g;
	Slength = norm(S);
	if Slength > delta
	  p = delta.*(S/Slength);
	  % disp(['C'])
	else
	  %Dogleg intersection 
	  [tau, ftau, info] = fzero (@(tau) norm(S+tau*(N-S))-delta,[0 1]);
	  if info == 1
	    p = S+tau*(N-S);
	  %  disp(['D'])
	  else
	    p = S; %If search for tau fails, use steepest descent step
	    % disp(['S'])
	  end
	end
      end    
      
    %Test prediction quality
    [candidate,cg,cH] = L(x+p,mu);
    if (current-m(current,g,H,p)) < mtol
      %The current iterate may be in an almost flat area (result of poor scaling)
      rho=0;
    else
      rho=(current-candidate)/(current-m(current,g,H,p));
    end

    %Accept or reject step and adapt trust region radius
    if rho >= etag
      x = x+p;
      delta = delta*gammae;
      
      best=candidate;
      solution = x;
      g=cg;
      gradlength = norm(g);
      H=cH;
    else
      if rho >= etaa
	x=x+p;
	
	best=candidate;
	solution = x;
	gradlength = norm(g);
	g=cg;
	H=cH;
      else 
	  delta = gammas*delta;
	  if delta < mindelta
	    delta = mindelta;
	  end
      end
    end  
    iter = iter + 1;

%      Traj=[Traj;x];
    Obj=[Obj;best];

    if diverge(Obj,solution) == 1
      divFlag=1;
      break;
    end
%      Delta = [Delta;delta];
  end  
  disp(['==================================================================']);
  mu
  objective(solution)
  solution
  disp(['C 1 = ' num2str(c1(solution))]);
  disp(['C 2 = ' num2str(c2(solution))]);

  %If method diverged, reinitialize from original point. Otherwise, use current solution
  if divFlag==1
    disp(['Diverges']);
  else
    xinit=solution;
  end
  
  divFlag=0;
  disp(['Using ' num2str(solution(1)) ' ' num2str(solution(2)) ' as initial point']);
  x=xinit;
  mu = mum*mu; %Increases penalty parameter  
  citer = mum*citer; % Increses maximum # of iterations
  if(citer > maxiter) %limits maximum # of iterations
    citer=maxiter;
  end
  piter = piter +1;
end

%  solution
%  best
